;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: ldata.lsp
;;;;    System: LTRE
;;;;   Version: 4.1 
;;;;    Author: Ken Forbus and Johan de Kleer
;;;;   Created: May 1, 1986
;;;;  Modified: everett on Fri Aug 2 9:34:57 1996
;;;;   Purpose: Database of datums and rules
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)

(defstruct (dbclass (:PRINT-FUNCTION dbclass-print-procedure))
  name    ; Corresponding symbol
  ltre    ; LTRE it is part of.
  facts   ; Associated facts
  rules)  ; Associated rules

(defun dbclass-print-procedure (r st ignore)
  (declare (ignore ignore))
  (format st "<Dbclass ~A>" (dbclass-name r)))

(defstruct (datum (:PRINT-FUNCTION datum-print-procedure))
   counter              ; Unique ID for easy lookup
   ltre                 ; The LTRE it is part of
   lisp-form            ; Expression for pattern-matching
   (tms-node nil)       ; Pointer into TMS
   dbclass              ; Dbclass of the corresponding pattern
   (plist nil))         ; local property list

(defun datum-print-procedure (d st ignore)
  (declare (ignore ignore)) 
   (if (datum-p d)
      (format st "<Datum ~D>" (datum-counter d))
      (format st "<Null Datum>")))

(defun alphalessp (x y) (string< (format nil "~A" x) (format nil "~A" y)))
(defun datum< (d1 d2) (< (datum-counter d1) (datum-counter d2)))
(defun form< (f1 f2) (datum< (referent f1 t) (referent f2 t)))


;;; ALLEGRO COMMON LISP FOR WINDOWS PATCH
;;; ----------------------------------------------------------------------------
;;; In ACLPC we can't hash nodes because equal failed when given numbers, 
;;; so we use form-equal, which turns out to be quite expensive when we 
;;; profile the ltre.  Grrrrr

(defvar *form-numerical-fuzz* 1.0e-9)

(defun form-equal (x y)
   ;;profiling says this function is expensive and frequently called
   (cond ((eq x y) t) 
         ((null x)
          (null y))
         ((and (consp x) (consp y))
          (and (form-equal (car x) (car y))
               (form-equal (cdr x) (cdr y))))
         ((and (numberp x) (numberp y))
          (< (abs (- x y)) *form-numerical-fuzz*))
         (t nil)))


;;; ADDING DATA
;;; ----------------------------------------------------------------------------
(defconstant *connective-list* '(:IMPLIES :AND :OR :IFF :NOT :TAXONOMY))

(defun simple-proposition? (x)
  (or (not (listp x))
      (not (member (car x) *connective-list*))))

(defun negated? (form)
  (and (listp form) (eq (car form) :NOT)))

(defun negated-proposition? (form)
 (and (negated? form)
      (simple-proposition? (cadr form))))

(defun unnegated (form)
  (if (negated? form) 
      (cadr form) form))

(defun assert! (fact just &optional (*LTRE* *LTRE*))
   "Install an incontrovertible fact into the database"
   (dbg-ltre :LTRE "Asserting ~S via ~S" fact just)
   ;; For assertions, simply install the clause.
   (add-formula (ltre-ltms *ltre*) (build-tms-formula fact *LTRE*) just))

(defun assume! (fact reason &optional (*LTRE* *LTRE*)
                  (assumption-error-checking? nil) &aux datum node)
   "Assume a fact that can later be retracted"
   (setq datum (referent (if (negated-proposition? fact) (cadr fact)
                            fact) t)
      node (datum-tms-node datum))
   (dbg-ltre :LTRE "  Assuming ~A via ~A." fact reason)
   (unless (or (negated-proposition? fact) (simple-proposition? fact))
      ;; Install clauses corresponding to the assumption, with proper
      ;; logical scoping
      (add-formula (ltre-ltms *ltre*)
         `(:IMPLIES ,node ,(build-tms-formula fact *LTRE*)) reason))
   (cond ((not (enabled-assumption? node))
          (convert-to-assumption node)
          (enable-assumption node
             (if (negated-proposition? fact) :FALSE :TRUE)
             reason))
         ((eq reason (tms-node-support node)))
         (assumption-error-checking?
          (error "Fact ~A assumed because of ~A assumed again because of ~A"
             (show-datum datum) (tms-node-support node) reason))
         (t nil))
   datum)

(defun already-assumed? (fact) (assumption? fact))

(defun quiet-assert! (fact *LTRE* &optional (just 'user))
  (without-contradiction-check (ltre-ltms *LTRE*)
		  (assert! fact *LTRE* just)))

(defmacro rassert! (fact &optional (just 'user))
  `(assert! ,(quotize fact) ,(quotize just)))

(defmacro rreferent! (fact &optional (just 'user))
   `(referent ,(quotize fact) ,(quotize just)))

(defun build-tms-formula (formula *LTRE*)
   (cond ((and (listp formula)
               (member (car formula) *connective-list*))
          (cons (car formula)
             (mapcar #'(lambda (form)
                          (build-tms-formula form *LTRE*)) (cdr formula))))
         (t (datum-tms-node (referent formula t)))))


;;; RETRACTION AND "BULK ASSUMPTIONS"
;;; ----------------------------------------------------------------------------
(defun retract! (fact just &optional (*LTRE* *LTRE*) (quiet? t)
                   &aux datum node)
   (setq datum (referent fact t)
      node (datum-tms-node datum))
   (cond ((not (tms-node-assumption? node))
          (unless quiet?
             (format t "~%~A isn't an assumption." (show-datum datum))))
         ((not (known-node? node))
          (unless quiet?
             (format T "~%The assumption ~A is not currently in." fact)))
         ((eq just (tms-node-support node))
          (dbg-ltre :LTRE "  Retracting ~A via ~A." fact just)
          (retract-assumption node))
         ((not quiet?)
          (format t "~%~A not source of assumption for ~A"
             just fact)))
   node)

(defmacro rretract! (fact &optional (just 'USER))
   "Use this with caution"
   `(retract! ,(quotize fact) ,(quotize just)))

(defun contradiction (losers &optional (informant :DECLARED-CONTRADICTION)
			     (*LTRE* *LTRE*)
			     &aux datum trues falses)
  ;; Here we re-cycle this procedure to indicate that a
  ;; conjunction of facts are contradictory.
  (dolist (fact losers
		(add-clause falses trues informant))
    (setq datum (referent fact t))
    (if (negated-proposition? fact)
	(push (datum-tms-node datum) falses)
      (push (datum-tms-node datum) trues))))

(defmacro assuming (facts-to-assume *LTRE* &body body)
  ;; A useful abstraction for user code.  
  `(with-assumptions
    (mapcar #'(lambda (fact &aux node datum)
		(setq datum (referent fact t)
		      node (datum-tms-node datum))
		(convert-to-assumption node)
		(cons node (if (negated-proposition? fact) :FALSE :TRUE)))
	    ,facts-to-assume)
    ,@ body))

(defmacro new-assuming (facts-to-assume informant *LTRE* &body body)
   `(let ((the-facts ,facts-to-assume)  ;; Make sure it's only evaluated once
          (the-informant ,informant))
       (unwind-protect
            (progn (dolist (fact the-facts)
                      (assume! fact the-informant *LTRE*))
               ,@ body)
          (dolist (fact the-facts)
             (cond ((and (known? fact)
                         (eq the-informant (informant-of fact)))
                    (retract! fact the-informant *ltre*))
                   (t (dbg-ltre :LTRE "   Not retracting ~A via ~A, ~A"
                         fact the-informant
                         (if (known? fact) :NOT-EQ :UNKNOWN))))))))


;;; DELETING DATA FOR FACT GARBAGE COLLECTION
;;; ----------------------------------------------------------------------------
(defun delete-datum (datum)
   (let ((datum-class (datum-dbclass datum)))
      (setf (datum-tms-node datum) nil)
      (setf (datum-plist datum) nil)
      (setf (dbclass-facts datum-class)
            (delete datum (dbclass-facts datum-class)))))
      

;;; DATABASE SYSTEM
;;; ----------------------------------------------------------------------------
(defun get-dbclass (fact &optional (*LTRE* *LTRE*) (error? t) &aux dbclass)
  (cond ((null fact)
	 (if error? (error "~% NIL can't be a dbclass.")
	   :UNKNOWN))
	((listp fact) 
	 (cond ((negated-proposition? fact)
		(get-dbclass (cadr fact) *LTRE* error?))
	       (t (get-dbclass (car fact) *LTRE* error?))))
	((variable? fact)
	 (cond ((boundp fact) (get-dbclass (symbol-value fact) *LTRE* error?))
	       (error? (error "~%Dbclass unbound: ~A" fact))
	       (t :UNKNOWN)))
        ((symbolp fact)
         (cond ((gethash fact (ltre-dbclass-table *LTRE*)))
               (t (setq dbclass
                     (make-dbclass 
                        :NAME fact :FACTS nil
                        :RULES nil :LTRE *LTRE*))
                  (setf (gethash fact (ltre-dbclass-table *LTRE*)) dbclass)
                  dbclass)))
	(t (error "Bad dbclass type: ~A" fact))))

(defun referent (fact &optional (virtual? nil) (*LTRE* *LTRE*))
   (if virtual? (insert fact) (referent1 fact)))

(defun referent1 (fact &aux form) ;; Could use seperate hash table
   (setq form (if (negated-proposition? fact) (cadr fact) fact))
   (dolist (candidate (dbclass-facts (get-dbclass fact *LTRE*)))
      (when (form-equal (datum-lisp-form candidate) form)
         (return candidate))))

(defun insert (fact &aux datum form)
   (setq datum (referent1 fact))
   (cond (datum (values datum t))
         (t (setq form (if (negated-proposition? fact) (cadr fact) fact))
            (setq datum (make-datum :COUNTER (incf (ltre-datum-counter *LTRE*))
                           :LTRE *LTRE*
                           :LISP-FORM form
                           :DBCLASS (get-dbclass form *LTRE*)))
            (setf (datum-tms-node datum)
                  (tms-create-node (ltre-ltms *LTRE*) datum))
            (push datum (dbclass-facts (datum-dbclass datum)))
            (try-rules datum) 
            (values datum nil))))

;; (defun fetch (pattern &optional (*LTRE* *LTRE*))
;;    (let ((bindings nil) (unifiers nil)
;;          (form (if (negated-proposition? pattern) (cadr pattern) pattern)))
;;       (dolist (candidate (get-candidates form *LTRE*) unifiers)
;;          (setq bindings (unify form (datum-lisp-form candidate) nil))
;;          (unless (eq bindings :FAIL)
;;             ;; Was (sublis bindings pattern)
;;             ;; Fetch is a big conser in GIZMO, this should help.
;;             ;; Assumes that nobody side-effects the pattern
;;             (push (datum-lisp-form candidate) unifiers)))))

;;; 6/7/96 JE modified fetch to take an optional filtering procedure as an arg 
;;; in order to move fetch-trues and fetch-falses into the ltre, from cyclepad.

(defun fetch (pattern &optional (*LTRE* *LTRE*) (filter #'cl-user::true))
   "Assumes pattern is never side-effected"
   (let ((unifiers nil)
         (form (if (negated-proposition? pattern) (cadr pattern) pattern)))
      (dolist (candidate (get-candidates form *LTRE*) unifiers)
         (let ((candidate-form (datum-lisp-form candidate)))
            (unless (or (eq (unify form candidate-form nil) :FAIL)
                        (not (funcall filter candidate-form)))
               (push candidate-form unifiers))))))

(defun fetch-trues (pattern &optional (*LTRE* *LTRE*))
   (fetch pattern *LTRE* 'true?))

(defun fetch-falses (pattern &optional (*LTRE* *LTRE*))
   (fetch pattern *LTRE* 'false?))

(defun fetch-unknowns (pattern &optional (*LTRE* *LTRE*))
   (fetch pattern *LTRE* 'unknown?))

(defun get-candidates (pattern *LTRE*)
  (dbclass-facts (get-dbclass pattern *LTRE*)))

(defun map-dbclass (proc &optional (*LTRE* *LTRE*))
  (maphash #'(lambda (name dbclass) (declare (ignore name))
	       (funcall proc dbclass))
	   (ltre-dbclass-table *LTRE*)))


;;; INTERFACE AND DISPLAY OF DATA
;;; ----------------------------------------------------------------------------
(defun true? (fact &optional (*LTRE* *LTRE*) &aux r)
  (when (setq r (referent fact nil))
    (if (negated-proposition? fact)
	(false-node? (datum-tms-node r))
	(true-node? (datum-tms-node r)))))

(defun false? (fact &optional (*LTRE* *LTRE*) &aux r)
  (when (setq r (referent fact nil))
    (if (negated-proposition? fact)
	(true-node? (datum-tms-node r))
	(false-node? (datum-tms-node r)))))

(defun known? (fact &optional (*LTRE* *LTRE*) &aux r)
  (when (setq r (referent fact nil))
    (known-node? (datum-tms-node r))))

(defun unknown? (fact &optional (*LTRE* *LTRE*) &aux r)
  (if (setq r (referent fact nil))
      (unknown-node? (datum-tms-node r))
      t))

(defun assumption? (fact &optional (*LTRE* *LTRE*) &aux r)
  (when (setq r (referent fact nil))
    (and (known-node? (datum-tms-node r))
         (enabled-assumption? (datum-tms-node r)))))

(defun assumption-source (fact &optional (*LTRE* *LTRE*) &aux r)
   (when (setq r (referent fact nil))
      ;;Assume it really is an assumption
      (tms-node-support (datum-tms-node r))))

(defun derived? (fact &optional (*ltre* *ltre*) &aux r)
  (when (setq r (referent fact nil))
	(and (known-node? (datum-tms-node r))
	     (typep (tms-node-support (datum-tms-node r))
		    'clause))))

(defun label-of (fact &optional (*ltre* *ltre*) &aux r)
  (when (setq r (referent fact nil))
	(tms-node-label (datum-tms-node r))))

(defun why? (fact &optional (*LTRE* *LTRE*) &aux r)
  (cond ((setq r (referent fact nil))
	 (why-node (datum-tms-node r)))
	(t (format t "~%~A not in database." fact))))

(defun get-tms-node (fact &optional (*LTRE* *LTRE*))
   (cond ((or (consp fact) (symbolp fact))
          (datum-tms-node (referent fact t)))
         ((numberp fact)
          (datum-tms-node (get-datum fact)))
         (t (error "Get-tms-node passed bad arg: ~A~%" fact))))

(defun view-node (d)
   (datum-lisp-form (tms-node-datum d)))

(defun signed-view-node (d)
   (if (false-node? d) (list :NOT (view-node d))
      (if (true-node? d) (view-node d)
         (error "SIGNED-VIEW-NODE requires knowing label: ~A" d))))

(defun show-datum (datum) 
   (when (datum-p datum)
      (format nil "~A" (datum-lisp-form datum))))
      
(defun make-node-string (node) 
   "This function passed in to LTMS"
   (show-datum (tms-node-datum node)))

(defun consequences (fact)
  (unless (known? fact) (return-from consequences nil))
  (show-node-consequences (get-tms-node fact)))

(defun explore (fact) (explore-network (get-tms-node fact)))


;;; GLOBAL INTERROGATIVES
;;; ----------------------------------------------------------------------------
(defun show-data (&optional (*LTRE* *LTRE*) (stream *standard-output*)
			    &aux counter)
  (setq counter 0)
  (format stream 
	  "~%~D facts total." (count-nodes))
  (map-dbclass
   #'(lambda (dbclass)
       (dolist (datum (dbclass-facts dbclass))
	 (incf counter)
	 (format stream "~%~A: ~A" (show-datum datum)
		 (cond ((true-node? (datum-tms-node datum)) "TRUE")
		       ((false-node? (datum-tms-node datum)) "FALSE")
		       (t "UNKNOWN"))))))
  counter)

(defun show-classes (&optional (*ltre* *ltre*) (stream *standard-output*)
			       &aux lines)
  (map-dbclass
   #'(lambda (dbclass)
       (push (format nil "~A: ~D facts, ~D rules"
		     (dbclass-name dbclass)
		     (length (dbclass-facts dbclass))
		     (length (dbclass-rules dbclass)))
	     lines)))
  (format stream "~%Summary of ~A:" (ltre-title *ltre*))
  (dolist (line (sort lines #'string<))
	  (format stream "~%  ~A" line)))

(defun show-class-facts (class &optional (stream *standard-output*))
   (dolist (fact (dbclass-facts (get-dbclass class)))
      (format stream "~&~A ~A~%"
         (tms-node-label (datum-tms-node fact))
         (datum-lisp-form fact))))

(defun scf (class &optional (stream *standard-output*))
  (show-class-facts class stream))
(export 'scf)

(defun get-datum (num &optional (*LTRE* *LTRE*))
  (maphash #'(lambda (key dbclass) 
	       (declare (ignore key))
	       (dolist (datum (dbclass-facts dbclass))
		       (when (= (datum-counter datum) num)
			     (return-from GET-DATUM datum))))
	   (ltre-dbclass-table *LTRE*)))

(defun get-clause (num &optional (*LTRE* *LTRE*))
   (map-over-clauses 
      #'(lambda (cl)
           (when (= (clause-index cl) num)
              (return-from get-clause cl)))))

(defun fetch-global (pattern &optional (status nil) (*ltre* *ltre*)
			     &aux results)
   (map-dbclass
   #'(lambda (dbclass)
       (dolist (datum (dbclass-facts dbclass))
	       (if (and (case status
			      (:TRUE (true-node? (datum-tms-node datum)))
			      (:FALSE (false-node? (datum-tms-node datum)))
			      (:KNOWN (known-node? (datum-tms-node datum)))
			      (:UNKNOWN (not (known-node? (datum-tms-node datum))))
			      (t t))
			(not (eq (unify (datum-lisp-form datum) pattern) :FAIL)))
		   (push (datum-lisp-form datum) results)))))
  results)

(defun map-over-datums (proc &optional (*ltre* *ltre*))
  (map-dbclass #'(lambda (dbclass)
                   (dolist (datum (dbclass-facts dbclass))
                     (funcall proc datum)))))

(defun map-over-facts (proc &optional (*ltre* *ltre*))
  (map-dbclass #'(lambda (dbclass)
                   (dolist (datum (dbclass-facts dbclass))
                     (funcall proc (datum-lisp-form datum))))))

(defun map-over-rules (proc &optional (*ltre* *ltre*))
  (map-dbclass #'(lambda (dbclass)
                   (dolist (rule (dbclass-rules dbclass))
                     (funcall proc rule)))))

(defun map-over-tms-nodes (proc &optional (*ltre* *ltre*))
  (map-dbclass #'(lambda (dbclass)
                   (dolist (datum (dbclass-facts dbclass))
                     (funcall proc (datum-tms-node datum))))))

(defun map-over-clauses (proc &optional (ltms (ltre-ltms *ltre*)))
   "Slow but safe way of scanning clauses"
   (let ((clause-mark (gensym)))
      (map-dbclass ;;install a new clause-mark
         #'(lambda (dbclass)
              (dolist (d (dbclass-facts dbclass))
                 (dolist (cl (tms-node-true-clauses (datum-tms-node d)))
                    (setf (clause-status cl) clause-mark))
                 (dolist (cl (tms-node-false-clauses (datum-tms-node d)))
                    (setf (clause-status cl) clause-mark)))))
      (map-dbclass ;;sweep through all clauses, applying proc once to each
         #'(lambda (dbclass)
              (dolist (d (dbclass-facts dbclass))
                 (dolist (cl (tms-node-true-clauses (datum-tms-node d)))
                    (when (eq (clause-status cl) clause-mark)
                       (funcall proc cl))
                    (setf (clause-status cl) nil))
                 (dolist (cl (tms-node-false-clauses (datum-tms-node d)))
                    (when (eq (clause-status cl) clause-mark)
                       (funcall proc cl))
                    (setf (clause-status cl) nil)))))
      nil))

(defun count-datums ()
   (let ((count 0))
      (map-over-tms-nodes #'(lambda (n)
                               (declare (ignore n)) (incf count)))
      (unless (= count (count-nodes))
         (error "Trouble in paradise: ~A nodes and ~A datums"
            (count-nodes) count))
      count))


;;; YET MORE USER HOOKS
;;; ----------------------------------------------------------------------------
;;; These are provided to better insulate other systems from LTMS internals

(defun antecedents-of (fact)
  (cond ((unknown? fact) nil) ;; Error?
        ((assumption? fact) nil)
        (t (mapcar #'view-node
                   (clause-antecedents
                    (tms-node-support (get-tms-node fact)))))))

(defun premise? (fact)
  (and (known? fact) (null (antecedents-of fact))))

(defun informant-of (fact)
  (cond ((unknown? fact) nil)
	((assumption? fact)
	 (tms-node-support (get-tms-node fact)))
	(t (compact-informant
	    (clause-informant
	     (tms-node-support
	      (get-tms-node fact)))))))

(defun assumptions-of (fact &optional (*LTRE* *LTRE*))
  (mapcar #'view-node 
	  (assumptions-of-node
	   (datum-tms-node (referent fact t)))))

(defun consequences-of (fact &optional (*LTRE* *LTRE*))
  (unless (known? fact) (return-from consequences-of nil))
  (mapcar #'view-node
	  (node-consequences
	   (get-tms-node fact))))


;;; MORE INTERROGATIVES
;;; ----------------------------------------------------------------------------
(defun clauses-of (fact &optional (*ltre* *ltre*))
  (let ((node (get-tms-node fact)))
    (when node
	  (node-show-clauses node))
    node))

(defun list-assumptions (&optional (*ltre* *ltre*))
  (remove-if-not #'assumption?
		 (fetch-global '?pattern :known)))

(defun show-assumptions (&optional (*ltre* *ltre*))
  (pprint (list-assumptions *ltre*)))

(defun retract!! (fact &optional (*ltre* *ltre*))
   "USE WITH CAUTION -- This retracts something no matter what"
   (retract! fact (informant-of fact)))

(defun force-assumption (fact reason &optional (*ltre* *ltre*))
   (when (premise? fact) (return-from force-assumption nil))
   (if (and (known? fact)
            (not (assumption? fact))) (retract!! fact))
   (assume! fact reason))

(defun nuke-assumptions (&optional (*ltre* *ltre*))
   (mapc #'retract!! (list-assumptions)))

;;;; Patch due to no longer automatically caching nodes in 
;;;; the TMS redundantly with LTRE

(defun explain (fact &optional (*ltre* *ltre*))
   (map-over-facts 
      #'(lambda (fact)
           (setf (tms-node-mark (get-tms-node fact))
                 nil)))
   (explain-node (get-tms-node fact)))

(defvar *line-count*)

(defun explain-node (node &aux *line-count*)
   (unless (eq (tms-node-label node) :UNKNOWN)
      (setq *line-count* 0)
      (explain-1 node)))

(defun explain-1 (node &aux antecedents)
  (cond ((tms-node-mark node))
	((enabled-assumption? node)
	 (format T "~%~3D ~15<~:[(:NOT ~A)~;~A~]~>~15<()~>   Assumption"
		 (incf *line-count*) (true-node? node) (node-string node))
	 (setf (tms-node-mark node) *line-count*))
	(t (setq antecedents
	     (mapcar #'explain-1 (clause-antecedents (tms-node-support node))))
	   (format T "~%~3D ~15<~:[(:NOT ~A)~;~A~]~> ~15<~A~>  "
		   (incf *line-count*) (true-node? node)
		   (node-string node) antecedents)
	   (pretty-print-clause (tms-node-support node))
	   (setf (tms-node-mark node) *line-count*))))

(defun explain-clause (clause &aux *line-count*)
  (setq *line-count* 0)
  (map-over-facts 
   #'(lambda (fact)
       (setf (tms-node-mark (get-tms-node fact))
	 nil)))
  (dolist (ante (clause-antecedents clause)) 
    (explain-1 ante))
  (format t "~%~3D " (incf *line-count*))
  (pretty-print-clause clause))

;;; ----------------------------------------------------------------------------
;;; END OF CODE    