;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: linter.lsp
;;;;    System: LTRE
;;;;   Version: 4.1
;;;;    Author: Ken Forbus, Johan de Kleer
;;;;   Created: May 1, 1986
;;;;  Modified: everett on Fri Aug 2 9:50:18 1996
;;;;   Purpose: LTRE Definitions
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)


;;; GLOBAL VARIABLE AND DATA STRUCTURE DEFINITIONS
;;; ----------------------------------------------------------------------------
(defvar *LTRE* nil "Points to current one--there may be several ltres in memory")

(defstruct (ltre (:PRINT-FUNCTION ltre-print-procedure))
   title                      ; Pretty name
   ltms                       ; Pointer to its LTMS
   (dbclass-table nil)        ; Hash table of dbclasses
   (gc-test #'default-gc-p)   ; If true of a datum-lisp-form, allows gc of form
   (datum-counter 0)          ; Unique ID for asserts
   (rule-counter 0)           ; Unique ID for rules
   (debugging nil)            ; Global switch for debugging information output
   (dbg-flags nil)            ; Flags that modulate which dbg info is printed
   (rule-queue :END-OF-Q)     ; LIFO queue of rules awaiting execution
   (rule-thread-dirty? nil)   ; Indicates ltre in midst of rule firing
   (rules-run 0))             ; Statistics

(defun ltre-print-procedure (l st ignore)
  (declare (ignore ignore))
  (format st "<LTRE: ~A>" (ltre-title l)))

;;; The binding of *ltre* is used inside rules and various macros to specify which
;;; LTRE a rule or fact should be stored in. The next few procedures encapsulate
;;; this choice

(defmacro with-LTRE (ltre &rest forms)
   `(let ((*LTRE* ,ltre)) ,@forms))

(defun in-LTRE (ltre)
   "Analogous to in-package"
   (setq *LTRE* ltre))


;;; MANAGING DEBUGGING INFORMATION ON THE LTRE
;;; ----------------------------------------------------------------------------
;;; Debugging information is provided by the macro dbg-ltre.
;;; These macros install debugging code in functions at compile time when the 
;;; parameter *strip-dbg-code* is nil.  Setting this parameter to T then 
;;; compiling or loading the ltre will result in uninstrumented code for the 
;;; runtime image, in which spurious debugging info would not be desired.
;;; The ltre-debugging slot of the ltre provides global control of debugging 
;;; information display, while the dbg-flags of the ltre enable modulation of
;;; the information display; these slots _must_ contain lists of keyword flags.
;;; Information is printed to *ltre-dbg-io*, which is by default bound to
;;; *debug-io*.  Lambda-binding them to a file-stream  or *trace-output* will
;;; serve to redirect debugging info.

(defparameter *strip-dbg-code* NIL "Remove all debugging instrumentation")

;;; We don't want debugging info showing up in runtime images
#+RUNTIME-SYSTEM (setq *strip-dbg-code* T)

(defvar *ltre-dbg-io* *debug-io* "Stream to which all debugging output goes")
(defvar *ltre-dbg-flags* nil "List of all debugging flags available")

(defmacro dbg-ltre (flag msg &rest msg-args)
   "Example: (dbg-ltre :RULE-Q \"Queueing ~~A\" rule)"
   (cond (*strip-dbg-code* (values))
         (t (pushnew flag *ltre-dbg-flags*)
            `(when (and (ltre-debugging *ltre*)
                        (or (member :ALL (ltre-dbg-flags *LTRE*) :test #'eq)
                            (member ,flag (ltre-dbg-flags *LTRE*) :test #'eq)))
                (format *ltre-dbg-io* ,msg ,@msg-args)))))


;;; CREATING AN LTRE
;;; ----------------------------------------------------------------------------
(defun create-ltre (title &key (gc-test #'default-gc-p) (setrule nil)
                      (debugging nil) (dbg-flags nil))
   (unless (listp dbg-flags) (setq dbg-flags (list dbg-flags)))
   (setq *ltre*
      (make-ltre
         :title title
         :gc-test gc-test
         :ltms (create-ltms (list :LTMS-OF title) 
                  :node-string 'make-node-string ;;see ldata.lsp
                  :cache-datums? nil)            ;; We do that
         :dbclass-table (make-hash-table :test #'eq)
         :debugging (consp dbg-flags)
         :dbg-flags (if (and debugging (null dbg-flags))
                         '(:ALL)
                       dbg-flags)))
   (when setrule
      (load-file *ltre-path* "setrule" :verbose nil))
   *ltre*)

(defun change-ltre (ltre &key (debugging nil debugging?))
   (when debugging? 
      (setf (ltre-debugging ltre) debugging)))
     

;;; RUNNING LTRE
;;; ----------------------------------------------------------------------------
(defun uassert! (fact &optional (just 'user) (*ltre* *ltre*))
  (assert! fact just) ;; Do internal operation
  (run-rules *ltre*))        ;; Run the rules

(defun uassume! (fact reason &optional (*ltre* *ltre*)) 
  (assume! fact reason)
  (run-rules *ltre*))

(defun run-forms (forms &optional (*LTRE* *LTRE*))
  (dolist (form forms) (eval form) (run-rules *ltre*)))

(defun run (&optional (*LTRE* *LTRE*)) ;; Toplevel driver function
    (format T "~%>>")
    (do ((form (read) (read)))
        ((member form '(quit stop exit abort)) nil)
        (format t "~%~A" (eval form))
        (run-rules)
        (format t "~%>>")))

(defun show-ltre (&optional (*LTRE* *LTRE*) (stream *standard-output*))
  (format stream "For LTRE ~A:" (ltre-title *LTRE*)) 
  (show-data *LTRE* stream) (show-rules *LTRE* stream))


;;; SOME DEBUGGING STUFF
;;; ----------------------------------------------------------------------------
(defun show-by-informant (informant &optional (*LTRE* *LTRE*)
				    &aux (count 0))
  (map-over-clauses
     #'(lambda (clause)
          (when (if (listp (clause-informant clause))
                   (eq (third (clause-informant clause)) informant)
                   (eq (clause-informant clause) informant))
             (incf count)
             (pprint (view-clause clause)))))
   count)

(defun view-clause (cl)
  (cons 'OR (mapcar #'(lambda (x)
			 (if (eq (cdr x) :FALSE) `(NOT ,(view-node (car x)))
			   (view-node (car x))))
		    (clause-literals cl))))

(defun statistics (&optional (*LTRE* *LTRE*) (stream *standard-output*))
   (format stream "~% For ~A: " (ltre-title *LTRE*))
   (format stream "~%  ~D datums, ~D rules, ~D executions."
      (count-datums) (count-rules)
      (ltre-rules-run *LTRE*))
   (if (rule-p (ltre-rule-queue *ltre*))
      (format stream "~%  ~D pending rules." (count-rules *ltre*))
      (format stream "~%  No pending rules."))
   (format stream "~%  ~D nodes, ~D clauses."
      (count-nodes)  (count-clauses))
   (format stream "~%  ~D nodes have been gc'd."
      (- (ltms-node-counter (ltre-ltms *ltre*)) (count-nodes)))
   (when (> (ltms-clause-counter (ltre-ltms *ltre*)) 0)
      (let ((min 10000) (max -10000) (sum 0))
         (map-over-clauses
            #'(lambda (c)
                 (let ((clen (length (clause-literals c))))
                    (if (> clen max) (setq max clen))
                    (if (< clen min) (setq min clen))
                    (incf sum clen))))
         (format stream
            "~%  Max|clause| = ~D; Min|clause| = ~D; Mean|clause| = ~D"
            max min (float (/ sum (ltms-clause-counter (ltre-ltms *ltre*))))))))

;;; ----------------------------------------------------------------------------
;;; END OF CODE